extern crate sdl2;

use sdl2::pixels::Color;

pub enum EgaColors {
    Black,
    Blue,
    Green,
    Cyan,
    Red,
    Magenta,
    Brown,
    LightGray,
    DarkGray,
    BrightBlue,
    BrightGreen,
    BrightCyan,
    BrightRed,
    BrightMagenta,
    BrightYellow,
    White,
}

impl TryFrom<u8> for EgaColors {
    type Error = &'static str;
    
    fn try_from(num: u8) -> Result<Self, Self::Error> {
        match num {
            0 => Ok(EgaColors::Black),
            1 => Ok(EgaColors::Blue),
            2 => Ok(EgaColors::Green),
            3 => Ok(EgaColors::Cyan),
            4 => Ok(EgaColors::Red),
            5 => Ok(EgaColors::Magenta),
            6 => Ok(EgaColors::Brown),
            7 => Ok(EgaColors::LightGray),
            8 => Ok(EgaColors::DarkGray),
            9 => Ok(EgaColors::BrightBlue),
            10 => Ok(EgaColors::BrightGreen),
            11 => Ok(EgaColors::BrightCyan),
            12 => Ok(EgaColors::BrightRed),
            13 => Ok(EgaColors::BrightMagenta),
            14 => Ok(EgaColors::BrightYellow),
            15 => Ok(EgaColors::White),
            _ => Err("EgaColors must be between 0 and 15 inclusive."),
        }
    }
}

impl Into<u8> for EgaColors {
    fn into(self) -> u8 {
        match self {
            EgaColors::Black => 0,
            EgaColors::Blue => 1,
            EgaColors::Green => 2,
            EgaColors::Cyan => 3,
            EgaColors::Red => 4,
            EgaColors::Magenta => 5,
            EgaColors::Brown => 6,
            EgaColors::LightGray => 7,
            EgaColors::DarkGray => 8,
            EgaColors::BrightBlue => 9,
            EgaColors::BrightGreen => 10,
            EgaColors::BrightCyan => 11,
            EgaColors::BrightRed => 12,
            EgaColors::BrightMagenta => 13,
            EgaColors::BrightYellow => 14,
            EgaColors::White => 15,
        }
    }
}

impl TryFrom<Color> for EgaColors {
    type Error = &'static str;

    fn try_from(color: Color) -> Result<Self, Self::Error> {
        match color {
            Color { r: 0x00, g: 0x00, b: 0x00, a: 0xff } => Ok(EgaColors::Black),
            Color { r: 0x00, g: 0x00, b: 0xaa, a: 0xff } => Ok(EgaColors::Blue),
            Color { r: 0x00, g: 0xaa, b: 0x00, a: 0xff } => Ok(EgaColors::Green),
            Color { r: 0x00, g: 0xaa, b: 0xaa, a: 0xff } => Ok(EgaColors::Cyan),
            Color { r: 0xaa, g: 0x00, b: 0x00, a: 0xff } => Ok(EgaColors::Red),
            Color { r: 0xaa, g: 0x00, b: 0xaa, a: 0xff } => Ok(EgaColors::Magenta),
            Color { r: 0xaa, g: 0x55, b: 0x00, a: 0xff } => Ok(EgaColors::Brown),
            Color { r: 0xaa, g: 0xaa, b: 0xaa, a: 0xff } => Ok(EgaColors::LightGray),
            Color { r: 0x55, g: 0x55, b: 0x55, a: 0xff } => Ok(EgaColors::DarkGray),
            Color { r: 0x55, g: 0x55, b: 0xff, a: 0xff } => Ok(EgaColors::BrightBlue),
            Color { r: 0x55, g: 0xff, b: 0x55, a: 0xff } => Ok(EgaColors::BrightGreen),
            Color { r: 0x55, g: 0xff, b: 0xff, a: 0xff } => Ok(EgaColors::BrightCyan),
            Color { r: 0xff, g: 0x55, b: 0x55, a: 0xff } => Ok(EgaColors::BrightRed),
            Color { r: 0xff, g: 0x55, b: 0xff, a: 0xff } => Ok(EgaColors::BrightMagenta),
            Color { r: 0xff, g: 0xff, b: 0x55, a: 0xff } => Ok(EgaColors::BrightYellow),
            Color { r: 0xff, g: 0xff, b: 0xff, a: 0xff } => Ok(EgaColors::White),
            _ => Err("The RGB color provided does not match an EgaColor."),
        }
    }
}

impl Into<Color> for EgaColors {
    fn into(self) -> Color {
        match self {
            EgaColors::Black => Color { r: 0x00, g: 0x00, b: 0x00, a: 0xff },
            EgaColors::Blue => Color { r: 0x00, g: 0x00, b: 0xaa, a: 0xff },
            EgaColors::Green => Color { r: 0x00, g: 0xaa, b: 0x00, a: 0xff },
            EgaColors::Cyan => Color { r: 0x00, g: 0xaa, b: 0xaa, a: 0xff },
            EgaColors::Red => Color { r: 0xaa, g: 0x00, b: 0x00, a: 0xff },
            EgaColors::Magenta => Color { r: 0xaa, g: 0x00, b: 0xaa, a: 0xff },
            EgaColors::Brown => Color { r: 0xaa, g: 0x55, b: 0x00, a: 0xff },
            EgaColors::LightGray => Color { r: 0xaa, g: 0xaa, b: 0xaa, a: 0xff },
            EgaColors::DarkGray => Color { r: 0x55, g: 0x55, b: 0x55, a: 0xff },
            EgaColors::BrightBlue => Color { r: 0x55, g: 0x55, b: 0xff, a: 0xff },
            EgaColors::BrightGreen => Color { r: 0x55, g: 0xff, b: 0x55, a: 0xff },
            EgaColors::BrightCyan => Color { r: 0x55, g: 0xff, b: 0xff, a: 0xff },
            EgaColors::BrightRed => Color { r: 0xff, g: 0x55, b: 0x55, a: 0xff },
            EgaColors::BrightMagenta => Color { r: 0xff, g: 0x55, b: 0xff, a: 0xff },
            EgaColors::BrightYellow => Color { r: 0xff, g: 0xff, b: 0x55, a: 0xff },
            EgaColors::White => Color { r: 0xff, g: 0xff, b: 0xff, a: 0xff },
        }
    }
}

impl EgaColors {
    pub fn u8_to_rgb(num: u8) -> Color {
        let col = match num {
            0 => Color::RGB(0x00, 0x00, 0x00),
            1 => Color::RGB(0x00, 0x00, 0xaa),
            2 => Color::RGB(0x00, 0xaa, 0x00),
            3 => Color::RGB(0x00, 0xaa, 0xaa),
            4 => Color::RGB(0xaa, 0x00, 0x00),
            5 => Color::RGB(0xaa, 0x00, 0xaa),
            6 => Color::RGB(0xaa, 0x55, 0x00),
            7 => Color::RGB(0xaa, 0xaa, 0xaa),
            8 => Color::RGB(0x55, 0x55, 0x55),
            9 => Color::RGB(0x55, 0x55, 0xff),
            10 => Color::RGB(0x55, 0xff, 0x55),
            11 => Color::RGB(0x55, 0xff, 0xff),
            12 => Color::RGB(0xff, 0x55, 0x55),
            13 => Color::RGB(0xff, 0x55, 0xff),
            14 => Color::RGB(0xff, 0xff, 0x55),
            15 => Color::RGB(0xff, 0xff, 0xff),
            _ => panic!("Unknown value for EgaColors: {}", num),
        };
        col
    }

    pub fn u8_to_textattr(num: u8) -> (bool, u8, u8) {
        // In EGA/VGA text mode, each screen character is represented by two
        // bytes. The low byte is the character, and the high byte is for
        // the attributes of the character:

        // +-------+-------+-------+-------+-------+-------+-------+-------+
        // |   7   |   6   |   5   |   4   |   3   |   2   |   1   |   0   |
        // +-------+-------+-------+-------+-------+-------+-------+-------+
        // | Blink |   Background Color    |       Foreground Color        |
        // +-------+-----------------------+-------------------------------+

        let blink = (num >> 7) != 0;
        let back = (num >> 4) & 7;
        let fore = num & 15;
        (blink, back, fore)
    }

    pub fn textattr_to_u8(blink: bool, background: u8, foreground: u8) -> u8 {
        let add_blink: u8 = match blink {
            false => 0,
            true => 128,
        };
        foreground + background * 16 + add_blink
    }
}